#ifndef HYPER_GENOM_TCLSERV_CLIENT_QUEUE_HH_
#define HYPER_GENOM_TCLSERV_CLIENT_QUEUE_HH_

#include <iostream>
#include <map>

#include <boost/asio.hpp>
#include <boost/bind.hpp>
#include <boost/function.hpp>
#include <boost/noncopyable.hpp>
#include <boost/tuple/tuple.hpp>

#include <tclserv_client/tclserv_client.h>

namespace hyper {
	namespace genom_model {
	class tclserv_client_queue : public boost::noncopyable {
		typedef boost::function<void (void)> cb_type;
		typedef std::map<ssize_t, cb_type> map_cb_type;

		map_cb_type map_cb;
		boost::asio::io_service& io_s;
		boost::asio::ip::tcp::socket s;
		TCLSERV_CLIENT_ID id;
		bool is_looping;

		void handle_read(const boost::system::error_code& e, std::size_t n)
		{
			if (!e) {
				/* Normally, handle_read triggers if there are some failure of
				 * if there are some data to read. Unfortunately, in some rare
				 * case, it seems to be triggered without error, but with an
				 * empty buffer, which leads to a starvation in tclserv_client_process_answer_msg
				 *
				 * To detect this specific case, call the bytes_readable
				 * command (ie FIONREAD on Unix) to compute the size of the tcp
				 * read buffer and enter in the real processing only if it is
				 * non 0.
				 */
				boost::asio::socket_base::bytes_readable command(true);
				s.io_control(command);
				std::size_t bytes_readable = command.get();
				if (bytes_readable > 0) {
					ssize_t rqstId;
					int err = tclserv_client_process_answer_msg(id, &rqstId);
					if ((err = tclserv_client_has_terminated(id, rqstId)) == 0) {
						// it can happens for ABORT msg for example
						map_cb_type::iterator it = map_cb.find(rqstId);
						if (it != map_cb.end()) {
							io_s.post(it->second);
							map_cb.erase(it);
						}
					}
				}

				if (! map_cb.empty()) 
					s.async_read_some(boost::asio::null_buffers(), 
							boost::bind(&tclserv_client_queue::handle_read, this, 
										boost::asio::placeholders::error,
										boost::asio::placeholders::bytes_transferred));
				else 
					is_looping = false;
			}
		}

		public:
			tclserv_client_queue(boost::asio::io_service& io_s, 
					const std::string& host,
					const std::string& port):
				io_s(io_s), s(io_s), is_looping(false)
			{
				id = tclserv_client_init_ex(host.c_str(), port.c_str(), false);
				assert(id != 0);
				// XXX, we can handle v6 too
				s.assign(boost::asio::ip::tcp::v4(), tclserv_client_get_answer_socket(id));
			}

			void enqueue(ssize_t rqstId, cb_type cb)
			{
				map_cb[rqstId] = cb;
				if (!is_looping) {
					s.async_read_some(boost::asio::null_buffers(), 
							boost::bind(&tclserv_client_queue::handle_read, this, 
								boost::asio::placeholders::error,
								boost::asio::placeholders::bytes_transferred));
				}
			}

			~tclserv_client_queue()
			{
				tclserv_client_destroy(&id);
			}

			TCLSERV_CLIENT_ID get_client_id() { return id; } 
	};
	}
}

#endif
